import { filter, Observable, Observer, Subject, Subscription } from "rxjs"
import { v4 as uuidv4 } from 'uuid'
import ConsoleLogger from "../utils/log.utils"
import { AdapterInterface, ClientObject, GeneralEvent, ReceiverAdapterInterface, TransmitterAdapterInterface, Transport, TransportServiceInterface } from "../interface/interface"
import { TransmitterAdapter } from "./adapter.transmitter";
import { ReceiverAdapter } from "./adapter.receiver";
import { AdapterManagerBase } from "../base/adapter.manager.base";

export class AdapterManager extends AdapterManagerBase {
    private console: ConsoleLogger = new ConsoleLogger(`AdapterManager`, ['managers'])

    constructor(event: Subject<GeneralEvent<any>>, browserEnv?: boolean) {
        super()
        this.browserEnv = browserEnv ?? false
        this.console.log({ message: `Contructing self...` })
        this.event = event

        this.connectToExistingTransport(this.event)
    }

    public subscribeForAdapters(): Observable<GeneralEvent<AdapterInterface>> {
        return new Observable((observer: Observer<GeneralEvent<AdapterInterface>>) => {
            const subscription: Subscription = this.event.pipe(
                filter((event: GeneralEvent<ClientObject>) => event.event === 'New Client')
            ).subscribe((event: GeneralEvent<ClientObject>) => {
                let adapters: AdapterInterface[] = this.instantiateAdapterComponents(event.data)
                if (adapters.length > 0) {
                    adapters.forEach((adapter: AdapterInterface) => {
                        observer.next({
                            id: uuidv4(),
                            type: `Adapter Event`,
                            event: `New Adapter`,
                            transport: event.data.transport,
                            date: new Date(),
                            data: adapter
                        })
                    })
                } else {
                    // handler for no adapters instantiated for this new client
                }
            })

            return () => {
                subscription.unsubscribe()
            }
        })
    }

    private connectToExistingTransport(event: Subject<GeneralEvent<any>>): void {
        this.event.pipe(
            filter(event => event.type === `Transport Event`),
            filter(event => event.event === `Available Transport`)
        ).subscribe((event: GeneralEvent<TransportServiceInterface[]>) => {
            let transportServices: TransportServiceInterface[] = event.data
            transportServices.forEach((transportService: TransportServiceInterface) => {
                this.updateTransportServicesRecord(transportService)
            })
        })
        this.event.next({
            id: uuidv4(),
            type: 'Adapter Event',
            event: `Adapter Manager Started`,
            date: new Date(),
            data: {
                event: event
            }
        })
    }

    private updateTransportServicesRecord(transportService: TransportServiceInterface): void {
        if (this.transportServiceArray.some(obj => obj.getInfo().transportServiceId === transportService.getInfo().transportServiceId)) {
            this.console.error({ message: `Transport Service Id: ${transportService.getInfo().transportServiceId} already existed.` })
        } else {
            this.transportServiceArray.push(transportService)
            this.console.log({ message: `Transport Service Id: ${transportService.getInfo().transportServiceId} has been added to adapter's record.` })
        }
    }

    private instantiateAdapterComponents(clientRef: ClientObject): AdapterInterface[] {
        let transportService = this.transportServiceArray.find(obj => obj.getInfo().transportServiceId === clientRef.transportServiceId)
        if (transportService) {
            let transmitterAdapter: TransmitterAdapterInterface = new TransmitterAdapter(clientRef.clientId, transportService.getInfo().transport, transportService)
            let receiverAdapter: ReceiverAdapterInterface = new ReceiverAdapter(clientRef.clientId, transportService.getInfo().transport, transportService)
            return [transmitterAdapter, receiverAdapter]
        } else {
            this.console.error({ message: `Transport Service id ${clientRef.transportServiceId} not found. Unable to instantiate adapters.` })
            return []
        }

    }
}


