import { filter, Observable, Observer, Subject, Unsubscribable } from "rxjs";
import { ActorInterface, ActorMessage, ActorProfile } from "../interface/actor.sample";

export class ActorBase<T> implements ActorInterface<T> {
    protected actorProfile!: ActorProfile
    protected incomingBus!: Subject<ActorMessage<T>>
    protected outgoingBus!: Subject<ActorMessage<T>>
    protected subscribers: ActorProfile[] = []

    /* so if one party call this to their respective target, they are essentially exposing their incomingBus so that both can pour messages through that channel.
     */
    public subscribe(subscriberProfile: ActorProfile, observer: Observer<ActorMessage<T>>): Unsubscribable {
        try {
            // Listen and update subscribers info.
            this.subscribers.push(subscriberProfile);

            // Filter out all other irrelevant messages and subscribe.
            const subscription = this.outgoingBus.pipe(
                filter((message: ActorMessage<T>) => message.actorProfile.actorId === subscriberProfile.actorId)
            ).subscribe({
                next: (message: ActorMessage<T>) => {
                    observer.next(message);
                },
                error: (err) => {
                    observer.error(err);
                },
            })

            // Return an Unsubscribable that cleans up both the subscription and subscriber info.
            return {
                unsubscribe: () => {
                    subscription.unsubscribe();
                    this.subscribers = this.subscribers.filter(
                        (sub) => sub.actorId !== subscriberProfile.actorId
                    );
                }
            };
        } catch (error) {
            observer.error(`Error during subscription: ${error}`);
            return {
                unsubscribe: () => {
                    // No-op in case of error
                }
            };
        }
    }



    protected setup(param: ActorProfile): void {
        throw new Error(`Method not implemented...`)
    }

    protected handleIncoming(message: Subject<T>): void {
        throw new Error(`Method not implemented...`)
    }

    protected handleOutgoing(message: Observable<T>): void {
        throw new Error(`Method not implemented...`)
    }

    protected checkSubscription(actorProfile: ActorProfile): boolean {
        return this.subscribers.some(actor => actor.actorId === actorProfile.actorId)
    }

}