import { filter, interval, map, Observable, Observer, Subject, take } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import { MessageTransmissionManager } from "../transmission/msg.transmission.manager";
import { WrappedMessage } from "../utils/message.ordering";
import ConsoleLogger from "../utils/log.utils";
import { FisMessage, GeneralEvent, TransmissionInterface, Transport, TransportMessage, TransportServiceInterface, TransportSet } from "../interface/interface";
import config from '../config/config.json';
import { startSocketServer } from "../utils/socket.utils";
import { WebsocketTransportService } from "../transport/websocket";
import { HttpTransportService } from "../transport/http";
class Supervisor {
    private console = new ConsoleLogger('Supervisor', ['base'])
    private clientIncomingMessage: Subject<FisMessage> = new Subject()
    private messageProducer!: MessageProducer
    private transmissionManager!: MessageTransmissionManager
    private event!: Subject<GeneralEvent<any>>
    private transmissionSets: TransmissionInterface[] = []
    private transportSet: TransportSet[] = []
    private transportServiceArray: TransportServiceInterface[] = []

    constructor() {
        this.event = new Subject<GeneralEvent<any>>()
        // set up all the transportServices
        this.sortTransportFromEnv(this.transportSet)
        this.transportSet.forEach(set => {
            this.setUpTransportService(set, this.event)
        })
        // so need them adapters now. But supervisor shouldn't be concerned, only messageTransmissionManager and ConnectionManager
        this.messageProducer = new MessageProducer(this.clientIncomingMessage)
        this.transmissionManager = new MessageTransmissionManager(this.event)
        this.startMessageTransmission(this.event)
    }

    private startMessageTransmission(transmissionEvent: Subject<GeneralEvent<TransmissionInterface>>): void {
        this.transmissionManager.subscribeForTransmission().pipe(
            filter(event => event.event == `New Transmission`)
        ).subscribe(event => {
            // update transmission record on every new transmission object instantiated
            this.transmissionSets.push(event.data as TransmissionInterface)
            // start message transmission for said transmission object instantiated
            this.handleClientActivity(event.data as TransmissionInterface)
        })
    }

    // only called once for each connected clients.
    private handleClientActivity(messageTransmission: TransmissionInterface): void {
        // start listening to incoming messages from this client
        messageTransmission.receiver.getIncoming().subscribe((event: GeneralEvent<TransportMessage>) => {
            let requestMessage: FisMessage = ((event.data as TransportMessage).payload as WrappedMessage).payload as FisMessage
            this.console.log({ message: `General Bus ${requestMessage?.header?.messageID ?? 'Not a message'}`, details: event }) // receiving end
            this.clientIncomingMessage.next(requestMessage)
            this.messageProducer.getOutgoingMessages().pipe(
                filter(message => message.header.messageID === requestMessage.header.messageID)
            ).subscribe(message => {
                messageTransmission.transmitter.emit(message)
            })
        })

        // to emulate general notification. Send every second
        this.messageProducer.getNotificationMessage().subscribe((message: FisMessage) => {
            messageTransmission.transmitter.emit(message)
        })
    }

    // Server to be set up as well as acquiring client information if needed. Like in the case for grpc and socket. Http not requ`ired.
    private setUpTransportService(transportSet: TransportSet, event: Subject<GeneralEvent<any>>, isClient?: boolean): void {
        try {
            let transportService: TransportServiceInterface = this.instantiateTransportService(transportSet.transport, event)
            this.transportServiceArray.push(transportService)
            if (transportService instanceof WebsocketTransportService) {
                this.console.log({ message: `Just Double Checking... this is websocket` })
                if (isClient) {
                    // please note this is subject to change depending on the UI environemnt. Angular has their own built in function to read json file based on Swopt-UI
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port);
                }
            } else if (transportService instanceof HttpTransportService) {
                this.console.log({ message: `Just Double Checking... this is http` })
                // Additional Http-specific setup if needed.
                if (isClient) {
                    transportService.startClient(config.connection.transmitter)
                } else {
                    transportService.startServer(transportSet.port)
                }
            }
        } catch (error) {
            this.console.error({ message: 'Fail to set transport. Error in setting up transport', details: error })
        }
    }

    private instantiateTransportService(transportType: Transport, event: Subject<GeneralEvent<any>>): TransportServiceInterface {
        if (transportType === 'Websocket') {
            return new WebsocketTransportService(event)
        }
        else if (transportType === 'Http') {
            return new HttpTransportService(event)
        } else {
            throw new Error(`No Transport Service Instantiated`)
        }
    }

    private sortTransportFromEnv(transportSet: TransportSet[]): void {
        let transportList: string[] = process.env.Transport?.split(',') || []
        let portList: number[] = (process.env.PORT?.split(',') || []).map(port => Number(port));
        transportList.forEach((transport, index) => {
            transportSet.push({ transport: transport, port: portList[index] } as unknown as TransportSet)
        })
        this.console.log({ message: 'TransportSetList', details: this.transportSet })
    }
}


class MessageProducer {
    private console = new ConsoleLogger('Message Producer', ['base'])
    private generalNotification: Subject<FisMessage> = new Subject()
    private incomingMessageBus!: Subject<FisMessage>
    private outgoingMessageBus: Subject<FisMessage> = new Subject()

    constructor(incomingMessageBus: Subject<FisMessage>) {
        this.console.log({ message: `Constructing Message Producer` })
        this.incomingMessageBus = incomingMessageBus

        this.generateNotifcation().subscribe(this.generalNotification)
        this.handleIncomingRequests(this.incomingMessageBus.asObservable(), this.outgoingMessageBus)
    }

    public getNotificationMessage(): Observable<FisMessage> {
        return this.generalNotification.asObservable()
    }

    public getOutgoingMessages(): Observable<FisMessage> {
        return this.outgoingMessageBus.asObservable()
    }

    // this is called no problem
    private handleIncomingRequests(requests: Observable<FisMessage>, outgoingMessageBus: Subject<FisMessage>): void {
        requests.subscribe((request: FisMessage) => {
            this.console.log({ message: `Generating response for new request ${request.header.messageID}` })
            this.generateMessage(request.header.messageID, 10).subscribe({
                next: message => outgoingMessageBus.next(message),
                error: error => this.console.log({ message: 'observer Error', details: error }),
                complete: () => {
                    outgoingMessageBus.next({
                        header: {
                            messageID: request.header.messageID,
                            messageName: `ResponseMessage`
                        },
                        data: `Complete`
                    } as FisMessage)
                }
            })
        })
    }

    private generateMessage(requestID: string, amount: number): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                take(amount), // Ensures only 'amount' messages are generated
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: requestID,
                            messageName: 'ResponseMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }

    private generateNotifcation(): Observable<FisMessage> {
        return new Observable((response: Observer<FisMessage>) => {
            const intervalMessageGeneration = interval(1000).pipe(
                map(() => {
                    const message: FisMessage = {
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Data`
                    };
                    return message;
                })
            );

            const subscription = intervalMessageGeneration.subscribe({
                next: message => response.next(message),
                error: error => response.error(error),
                complete: () => {
                    response.next({
                        header: {
                            messageID: uuidv4(),
                            messageName: 'NotificationMessage'
                        },
                        data: `Complete`
                    });
                    response.complete();
                }
            });

            // Ensure cleanup on unsubscribe
            return () => subscription.unsubscribe();
        });
    }
}

let supervisor = new Supervisor()
