import { Express } from 'express';
import { filter, Observable, Subject, Subscription, take } from "rxjs";
import { v4 as uuidv4 } from 'uuid'
import config from '../config/config.json';
import { handleClientHttpConnection, handleHttpClient, initiateClientToServer, startHttpServer } from "../utils/http.utils";
import { WrappedMessage } from '../utils/message.ordering';
import { error } from 'console';
import axios, { AxiosError } from 'axios';
import ConsoleLogger from '../utils/log.utils';
import { ClientObject, GeneralEvent, Transport, TransportMessage, TransportServiceInterface, TransportServiceProfile } from '../interface/interface';

export class HttpTransportService implements TransportServiceInterface {
    private console: ConsoleLogger = new ConsoleLogger(`HttpTransportService`, ['transport'])
    private baseUrl!: string;
    private info!: TransportServiceProfile
    private connectedHttpServer: ConnectedHttpServer[] = [] // to allow the possibility of having to communicate with multiple servers as a client
    private connectedHttpClients: ConnectedHttpClient[] = [] // to keep track of the all the clients that are connected
    transportEvent!: Subject<GeneralEvent<any>>

    constructor(event: Subject<GeneralEvent<any>>) {
        this.info = {
            transportServiceId: uuidv4(),
            transport: 'Http'
        }
        this.baseUrl = config.connection.transmitter
        this.transportEvent = event
    }

    public subscribeForTransportEvent(): Observable<GeneralEvent<any>> {
        throw new Error('Method not implemented.');
    }

    public getInfo(): TransportServiceProfile {
        return this.info
    }

    public emit(message: TransportMessage): void {
        let clientObj: ConnectedHttpClient | undefined = this.connectedHttpClients.find(obj => obj.clientId == message.target)
        let serverObj: ConnectedHttpServer | undefined = this.connectedHttpServer.find(obj => obj.clientId === message.target)

        // for server usage
        if (clientObj && clientObj.connectionState.getValue() == 'ONLINE') {
            clientObj.responseStream.next(message.payload as WrappedMessage)
        }
        // for client usage 
        if (serverObj) {
            if (serverObj.connectionState.getValue() == 'ONLINE') {
                axios.post(`${this.baseUrl}message`, message.payload, {
                    headers: { 'Content-Type': 'application/json' },
                }).then((response) => {
                    this.console.log({ message: `Response From Server: ${response.data}`, details: response.status });
                }).catch((error: AxiosError) => {
                    console.error('HTTP emit error:', error.code);
                    this.transportEvent.next({
                        id: uuidv4(),
                        type: 'Transport Event',
                        event: 'Re-Flush',
                        date: new Date(),
                        data: message,
                        transport: 'Http'
                    } as GeneralEvent<TransportMessage>)
                });
            } else {
                this.console.error({ message: `Target Server is offline: Reflusing message ${(message.payload as WrappedMessage).thisMessageID}` });
                this.transportEvent.next({
                    id: uuidv4(),
                    type: 'Transport Event',
                    event: 'Re-Flush',
                    date: new Date(),
                    data: message,
                    transport: 'Http'
                } as GeneralEvent<TransportMessage>)
            }
        }

    }

    public startServer(port: number): void {
        startHttpServer(port).subscribe({
            next: (client: ConnectedHttpClient) => {
                handleHttpClient(this.info.transportServiceId, client, this.connectedHttpClients).subscribe({
                    next: event => this.transportEvent.next(event),
                    error: error => console.error(error),
                    complete: () => (`Client ${client.clientId} disconnected...`)
                })
            },
            error: error => this.console.error({ message: 'Observer Error', details: error }),
            complete: () => this.console.log({ message: '...' })
        })
    }

    public startClient(url: string, receiverProfileInfo?: ConnectedHttpServer | undefined): void {
        initiateClientToServer(this.info.transportServiceId, url, this.transportEvent, this.connectedHttpServer, receiverProfileInfo).then((connectedHttpServer: ConnectedHttpServer) => {
            handleClientHttpConnection(url, connectedHttpServer).subscribe({
                next: event => this.transportEvent.next(event),
                error: error => this.console.log({ message: `Observer Error`, details: error }),
                complete: () => {
                    this.console.log({ message: `Re-connecting with server` })
                    this.startClient(url, connectedHttpServer)
                }
            })
        }).catch((error: { error: AxiosError, objRef: ConnectedHttpServer | undefined }) => {
            this.console.error({ message: `HttpTransport ERROR: <InitiateClientToServerFailure>`, details: error.error.code })
            setTimeout(() => {
                this.startClient(url, error.objRef)
            }, 3000); // Retry with delay
        })
    }

}


export interface ConnectedHttpClient extends ClientObject {
    instance: Express
    responseStream: Subject<WrappedMessage>
}

export interface ConnectedHttpServer extends ClientObject {
}