import { BehaviorSubject, Observable, Observer, Subject } from 'rxjs';
import { createServer } from 'http';
import { Server, Socket as SocketForConnectedClient } from 'socket.io';
import { io, Socket as SocketForConnectedServer } from 'socket.io-client';
import * as fs from 'fs'
import { v4 as uuidv4 } from 'uuid'
import { ConnectedSocketClient, ConnectedSocketServer } from '../transport/websocket';
import ConsoleLogger from './log.utils';
import path from 'path';
import { ConnectionState, GeneralEvent, TransportMessage } from '../interface/interface';
import { addClientToDB, checkIfClientExists, checkOwnClientInfo, writeFile } from './general.utils';
const console: ConsoleLogger = new ConsoleLogger(`SocketUtils`, ['transport'])

export function startSocketServer(port: number): Observable<SocketForConnectedClient> {
    return new Observable((observer) => {
        try {
            console.log({ message: `Socket Server ${port} Started....` })
            let httpServer = createServer();
            let socketServer = new Server(httpServer)

            // something wrong here
            socketServer.on('connection', (socket) => {
                observer.next(socket)
            })

            socketServer.engine.on("connection_error", (err) => {
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.req })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.code })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.message })
                console.log({ message: `Socket Server ${port} Connection Error`, details: err.context })
            });

            // Start the HTTP server on 127.0.0.1 with the given port
            httpServer.listen(port, '0.0.0.0', () => {
                console.log({ message: `Socket server listening on ${port}` });
            });
        } catch (error) {
            observer.error(error);
        }
    })
}

export async function startClientSocketConnection(serverUrl: string): Promise<SocketForConnectedServer> {
    return new Promise((resolve, reject) => {
        try {
            // let clientSocket = io(serverUrl)
            let clientSocket: SocketForConnectedServer = io(serverUrl, {
                reconnection: true,              // Enable automatic reconnections
                reconnectionAttempts: 1000,       // Retry up to 10 times
                reconnectionDelay: 500,          // Start with a 500ms delay
                reconnectionDelayMax: 10000,     // Delay can grow to a max of 10 seconds
                randomizationFactor: 0.3,
            })
            resolve(clientSocket)
        }
        catch (error) {
            reject(error)
        }
    })
}

// After establishing connection to the server, set up the credentials, confirm whether or not if there's any credentials, if not ask for one from the server
export function handleClientSocketConnection(transportServiceId: string, socket: SocketForConnectedServer, serversConnected: ConnectedSocketServer[]): Observable<GeneralEvent<any>> {
    return new Observable((eventNotification: Observer<GeneralEvent<any>>) => {
        let buffer: any[] = []
        let receiverProfileInfo!: ConnectedSocketServer

        // Listen for a connection event
        socket.on('connect', () => {
            console.log({ message: `Connected to the server ${socket.id} ` })
            if (receiverProfileInfo?.clientId) {
                checkOwnClientInfo(receiverProfileInfo.clientId).then((profile: { id: string }) => {
                    socket.emit('profile', {
                        name: 'Old Client',
                        data: profile
                    })
                }).catch((error) => {
                    socket.emit('profile', {
                        name: 'New Client',
                        data: null
                    })
                })
            } else {
                socket.emit('profile', { 
                    name: 'New Client',
                    data: null
                })
            }
        });

        // Listen for messages from the server. Generally here's the responses
        socket.on('message', (msg: any) => {
            // console.log(`Websocket Client Transport Receieve Msg`, msg)
            if (receiverProfileInfo) {
                // publish to event
                eventNotification.next({
                    id: uuidv4(),
                    type: `Transport Event`,
                    event: 'New Message',
                    date: new Date(),
                    data: {
                        id: uuidv4(),
                        dateCreated: new Date(),
                        transport: `Websocket`,
                        target: receiverProfileInfo.clientId,
                        payload: msg
                    } as TransportMessage
                })
            } else {
                // Do nothing. just store in local array first. Cannot process without information. but then again, don['t need information if acting as client
                // but for consistency sake, will impose the standard 
                buffer.push(msg) // store locally for now
            }
        })

        socket.on('profile', (data: { name: string, message: any }) => {
            if (data.name == 'New Profile') {
                console.log({ message: `Assigned client Name: ${data.message.id}` })
                // Update websocket instance record
                receiverProfileInfo = {
                    clientId: data.message.id,
                    dateCreated: new Date(),
                    socketInstance: socket,
                    connectionState: new BehaviorSubject<ConnectionState>(`ONLINE`),
                    transport: 'Websocket',
                    transportServiceId: transportServiceId
                }
                writeFile(data.message as ConnectedSocketServer, (data.message as ConnectedSocketServer).clientId).then(() => {
                    // broadcast event to allow transmission manager to instantiate transmission components
                    eventNotification.next({
                        id: uuidv4(),
                        type: 'Transport Event',
                        event: `New Server`,
                        date: new Date(),
                        data: {
                            clientId: (data.message as ConnectedSocketServer).clientId,
                            message: `New Websocket Channel ${(data.message as ConnectedSocketServer).clientId} established.`
                        },
                        transport: 'Websocket'
                    })
                    // broadcast event to allow retransmission to relase buffered messages
                    eventNotification.next({
                        id: uuidv4(),
                        type: 'Transport Event',
                        event: `Server Connected`,
                        date: new Date(),
                        data: {
                            clientId: (data.message as ConnectedSocketServer).clientId,
                            message: `Server ${(data.message as ConnectedSocketServer).clientId} connected and ready to go.`
                        },
                        transport: 'Websocket'
                    })
                }).catch((error) => { }) // do nothing at the moment. 
                serversConnected.push(receiverProfileInfo)
            }
            if (data.name == 'Adjusted Profile') {
                console.log({ message: `Adjusted client Name: ${(data.message as ConnectedSocketServer).clientId}` })
                // Update websocket instance record
                let clientObj: ConnectedSocketServer | undefined = serversConnected.find(obj => obj.clientId === data.message.id)
                if (clientObj) {
                    clientObj.socketInstance = socket
                    clientObj.connectionState.next('ONLINE')
                    console.log({
                        message: `Just to make sure they are pointed accurately: This should be ONLINE: ${receiverProfileInfo.connectionState.getValue()} !! Id match? ${receiverProfileInfo.clientId == clientObj.clientId ? true : false} && compare ${clientObj.clientId}`,
                    })
                    // broadcast event to allow retransmission to release buffer
                    eventNotification.next({
                        id: uuidv4(),
                        type: `Transport Event`,
                        event: 'Server Connected',
                        date: new Date(),
                        data: {
                            clientId: (data.message as ConnectedSocketServer).clientId,
                            message: `Existing Websocket Channel ${(data.message as ConnectedSocketServer).clientId} re-established.`
                        },
                        transport: 'Websocket'
                    })
                }
            }
            if (data.name == 'Error') {
                console.log({ message: `Server cannot find credentials`, details: data.message })
                // logic to request for new credentials
                setTimeout(() => {
                    socket.emit('profile', {
                        name: 'New Client',
                        data: null
                    })
                }, 2000)
            }
        })

        // Handle disconnection
        socket.on('disconnect', () => {
            console.log({ message: `Socket Server ${receiverProfileInfo.clientId} Disconnected` })
            if (receiverProfileInfo) {
                eventNotification.next({
                    id: uuidv4(),
                    type: `Transport Event`,
                    event: `Server Disconnected`,
                    date: new Date(),
                    data: {
                        clientId: receiverProfileInfo.clientId,
                        message: `Server for Channel ${receiverProfileInfo.clientId} disconnected.`
                    },
                    transport: `Websocket`
                })
                receiverProfileInfo.connectionState.next(`OFFLINE`)
            }
        });
    })
}

// For SERVER Usage: set up socket listeners to start listening for different events
export function handleNewSocketClient(transportServiceId: string, socket: SocketForConnectedClient, connectedClientSocket: ConnectedSocketClient[]): Observable<GeneralEvent<any>> {
    return new Observable((event: Observer<GeneralEvent<any>>) => {
        console.log({ message: `Setting up listeners for socket:${socket.id}` })
        // returns the socket client instance 
        // listen to receiver's initiotion first before assigning 'credentials'
        socket.on(`profile`, (message: { name: string, data: any }) => {
            if (message.name == 'New Client') {
                let clientInstance: ConnectedSocketClient = {
                    clientId: uuidv4(), // client should only be assigned at this level. And is passed around for reference pointing
                    dateCreated: new Date(),
                    socketInstance: socket,
                    connectionState: new BehaviorSubject<ConnectionState>(`OFFLINE`),
                    transport: 'Websocket',
                    transportServiceId: transportServiceId
                }
                // send to receiver for reference
                socket.emit('profile', {
                    name: `New Profile`, message: { id: clientInstance.clientId }
                })
                // publish first event notification
                event.next({
                    id: uuidv4(),
                    type: 'Transport Event',
                    event: `New Client`,
                    date: new Date(),
                    data: {
                        clientId: clientInstance.clientId,
                        message: `New Socket Client Connected. Adapter ID assigned: ${clientInstance.clientId}`,
                        payload: clientInstance
                    },
                    transport: 'Websocket'
                })
                // Update connected clientInstance info to adapter
                connectedClientSocket.push(clientInstance)
                addClientToDB(clientInstance)
                startListening(socket, clientInstance, event)
            } else {
                // update first
                let clientInstance: ConnectedSocketClient | undefined
                if (connectedClientSocket.length > 0) {
                    clientInstance = connectedClientSocket.find(obj => obj.clientId === message.data.id)
                    handleFoundClient(clientInstance)
                } else {
                    // for the case server itself got shit down or something
                    checkIfClientExists(message.data.id).then((client: ConnectedSocketClient) => {
                        clientInstance = client
                        handleFoundClient(clientInstance)
                    }).catch(error => {
                        console.log({ message: `Promise Error`, details: error })
                    })
                }
                function handleFoundClient(clientInstance: ConnectedSocketClient | undefined) {
                    if (clientInstance) {
                        console.log({ message: `Socket Client ${clientInstance.clientId} Found` })
                        socket.emit('profile', { name: 'Adjusted Profile', message: { id: clientInstance.clientId } })
                        // replace socket instance since the previous has been terminated
                        clientInstance.socketInstance = socket
                        // need to start listening again, because it's assigned a different socket instance this time round
                        startListening(socket, clientInstance, event, true)
                    } else {
                        console.log({ message: `Profile Not Found` })
                        socket.emit('profile', { name: 'Error', message: 'Receiver Profile Not found' })
                    }
                }
            }
        })
    })
}


// this is for server usage only
export function startListening(socket: SocketForConnectedClient, client: ConnectedSocketClient, eventListener: Observer<GeneralEvent<any>>, oldClient?: boolean): void {
    // notify it's associated retransmission to start releaseing buffer
    eventListener.next({
        id: uuidv4(),
        type: 'Transport Event',
        event: oldClient ? 'Client Re-connected' : `Client Connected`,
        date: new Date(),
        data: {
            clientId: client.clientId,
            message: `Socket Client ${oldClient ? `Re-Connected` : `Connected`}. Adapter ID assigned: ${client.clientId}`,
            payloclientId: client
        },
        transport: 'Websocket'
    })
    // Resume operation
    // some explanation here. For the case where the server reads from the DB, no need to terminate subject, since all instances would be destroyed alongside the server shut down. This case is specificd only when there's a need to read from local file
    if (!client.connectionState) {
        client.connectionState = new BehaviorSubject<ConnectionState>(`ONLINE`)
    } else {
        client.connectionState.next(`ONLINE`)
    }

    /* Generally, we don't need this unless in the case of being the receiver */
    socket.on('message', (message: any) => {
        console.log({ message: `Message from client ${client.clientId}`, details: message })
        eventListener.next({
            id: uuidv4(),
            type: 'Transport Event',
            event: 'New Message',
            date: new Date(),
            data: {
                id: uuidv4(),
                dateCreated: new Date(),
                transport: `Websocket`,
                target: client.clientId, // this ref to be associated with the client/channel
                payload: message
            } as TransportMessage
        })
    })

    socket.on('disconnect', () => {
        eventListener.next({
            id: uuidv4(),
            type: `Transport Event`,
            event: 'Client Disconnected',
            date: new Date(),
            data: {
                clientId: client.clientId,
                message: '',
                payload: {
                    time: new Date()
                }
            },
            transport: 'Websocket'
        })
        eventListener.error(`Client ${client.clientId} disconnected. Terminating this observable event for this client socket...`)
        eventListener.complete()
    })
}
